from .basic import Chatbot


class BoseBot(Chatbot):
    """Represents an agent in Navigator.
        https://eminent-kiwi-trusty.ngrok-free.app/api/bose/messages
        Each agent has a name, a role, a goal, a backstory,
        and an optional language model (llm).
    """
    name: str = 'BoseBot'
    company: str = 'T-ROC Global'
    company_website: str = 'https://www.trocglobal.com'
    contact_information = 'communications@trocglobal.com'
    contact_form = 'https://bose.trocdigital.io/bose/bose_ticketing_system'
    role: str = 'Bose Sound Systems Expert Technician and Consultant.'
    goal = 'Bring useful information to Bose Technicians and Consultants.'
    specialty_area = 'Bose endcap displays that are used to showcase Bose products'
