from .abstract import AbstractChatbot
from .basic import Chatbot
from .asktroc import AskTROC
from .hragents import HRAgent
from .bose import BoseBot
from .odoo import OddieBot
from .cody import Cody
