from .dir import load_directory
from .pdf import PDFLoader
from .web import WebLoader
from .youtube import YoutubeLoader
from .vimeo import VimeoLoader
from .word import MSWordLoader
from .ppt import PPTXLoader
from .repo import RepositoryLoader
from .github import GithubLoader
from .json import JSONLoader
from .excel import ExcelLoader
from .web_base import WebBaseLoader
from .pdfmark import PDFMarkdownLoader
from .pdfimages import PDFImageLoader
from .pdftables import PDFTablesLoader
from .pdfchapters import PDFChapterLoader
from .txt import TXTLoader
from .qa import QAFileLoader
from .rtd import ReadTheDocsLoader
from .videolocal import VideoLocalLoader
